/*
**  Copyright 2001-2003  University of Illinois Board of Trustees
**  Copyright 2001-2003  Mark D. Roth
**  All rights reserved. 
**
**  pam_dbm.c - PAM module for DBM file authentication
**
**  Mark D. Roth <roth@uiuc.edu>
**  Campus Information Technologies and Educational Services
**  University of Illinois at Urbana-Champaign
*/

#include <config.h>

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/param.h>
#include <syslog.h>

#ifdef STDC_HEADERS
# include <string.h>
# include <stdlib.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_CRYPT_H
# include <crypt.h>
#endif

#include <security/pam_appl.h>
#include <security/pam_modules.h>

#if HAVE_NDBM_H
# include <ndbm.h>
#elif HAVE_GDBM_NDBM_H
# include <gdbm/ndbm.h>
#endif

#include <compat.h>


#ifndef LOG_AUTHPRIV
# define LOG_AUTHPRIV LOG_AUTH
#endif


/*
** dbm_lookup() - look up key in DBM file
** returns:
**	0			success
**	-1 (and sets errno)	failure
*/
static int
dbm_lookup(char *dbm_file, char *key, char *buf, size_t buflen)
{
	DBM *dbm;
	datum keydat, valdat;

	dbm = dbm_open(dbm_file, O_RDONLY, 0);
	if (dbm == NULL)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_dbm_authenticate: dbm_open(\"%s\") failed",
		       dbm_file);
		return -1;
	}

	keydat.dptr = key;
	keydat.dsize = strlen(key) + 1;
	valdat = dbm_fetch(dbm, keydat);
	if (valdat.dptr == NULL)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_dbm_authenticate: dbm_fetch(\"%s\") failed", key);
		return -1;
	}

	dbm_close(dbm);

	strlcpy(buf, valdat.dptr, buflen);
	return 0;
}


int
pam_sm_authenticate(pam_handle_t *pamh, int flags,
		    int argc, const char *argv[])
{
	char *user, *service, *gencryptpw;
	char realcryptpw[1024];
	struct pam_message prompt = { PAM_PROMPT_ECHO_OFF, NULL };
	struct pam_message *msg[1];
	struct pam_conv *conv;
	struct pam_response *response = NULL;

#ifdef DEBUG
	syslog(LOG_AUTHPRIV | LOG_DEBUG,
	       "==> pam_dbm_authenticate(pamh=0x%lx, flags=%d)", pamh, flags);
#endif

	if (argc != 1)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_dbm_authenticate: argc=%d - expecting 1", argc);
		return PAM_AUTHINFO_UNAVAIL;
	}

	if (pam_get_user(pamh, (char **)&user, "login: ") != PAM_SUCCESS)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_dbm_authenticate: pam_get_user failed");
		return PAM_USER_UNKNOWN;
	}

	if (pam_get_item(pamh, PAM_SERVICE, (void **)&service) != PAM_SUCCESS)
		service = "*";

	prompt.msg = "Password: ";
	msg[0] = &prompt;
	if (pam_get_item(pamh, PAM_CONV, (void **)&conv) != PAM_SUCCESS ||
	    conv->conv(1, (struct pam_message **)msg,
		       &response, conv->appdata_ptr) != PAM_SUCCESS ||
	    response == NULL || response[0].resp == NULL)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_dbm_authenticate: cannot get user response");
		return PAM_AUTHINFO_UNAVAIL;
	}

	if (dbm_lookup((char *)argv[0], user,
		       realcryptpw, sizeof(realcryptpw)) == -1)
		return PAM_AUTHINFO_UNAVAIL;

	gencryptpw = crypt(response[0].resp, realcryptpw);
	if (gencryptpw == NULL)
	{
		syslog(LOG_AUTHPRIV | LOG_ERR,
		       "pam_dbm_authenticate: crypt(): %m");
		return PAM_AUTHINFO_UNAVAIL;
	}

	if (strcmp(realcryptpw, gencryptpw) != 0)
		return PAM_AUTH_ERR;

	return PAM_SUCCESS;
}


int
pam_sm_setcred(pam_handle_t *pamh, int flags, int argc, const char *argv[])
{
#ifdef DEBUG
	syslog(LOG_AUTHPRIV | LOG_ERR,
	       "pam_dbm_setcred: not implemented");
#endif
	return PAM_SUCCESS;
}


int
pam_sm_acct_mgmt(pam_handle_t *pamh, int flags, int argc, const char *argv[])
{
	syslog(LOG_AUTHPRIV | LOG_ERR,
	       "pam_dbm_acct_mgmt: not implemented");
	return PAM_IGNORE;
}


int
pam_sm_open_session(pam_handle_t *pamh, int flags,
		    int argc, const char *argv[])
{
	syslog(LOG_AUTHPRIV | LOG_ERR,
	       "pam_dbm_open_session: not implemented");
	return PAM_IGNORE;
}


int
pam_sm_close_session(pam_handle_t *pamh, int flags,
		     int argc, const char *argv[])
{
	syslog(LOG_AUTHPRIV | LOG_ERR,
	       "pam_dbm_close_session: not implemented");
	return PAM_IGNORE;
}


int
pam_sm_chauthtok(pam_handle_t *pamh, int flags, int argc, const char *argv[])
{
	syslog(LOG_AUTHPRIV | LOG_ERR,
	       "pam_dbm_chauthtok: not implemented");
	return PAM_IGNORE;
}


